open Printf;;

type color =
  | Red | Green | Blue | Cyan | Magenta | Yellow | Black | White | Initial
;;

let string_of_color = function
 | Red -> "r"
 | Green -> "g"
 | Blue -> "b"
 | Cyan -> "c"
 | Magenta -> "m"
 | Yellow -> "y"
 | Black -> "k"
 | White -> "w"
 | Initial -> assert false
;;

type tag_type =
 | B
 | EM
 | I
 | PL
 | S
 | TT
 | U
 | Size of int
 | Color of color
;;

let string_of_tag = function
 | U -> "U"
 | S -> "S"
 | EM -> "EM"
 | I -> "I"
 | B -> "B"
 | TT -> "TT"
 | PL -> "PL"
 | Size (-1) -> assert false
 | Size int -> string_of_int int
 | Color color -> string_of_color color
;;

type lexem =
   | Text of string
   | Space of string
   | Open of tag_type
   | Close of tag_type
   | EOF
;;

let column = ref 0;;

let print_lexeme l =
  if !column >= 40 then begin
    printf "\n";
    column := 0;
  end;
  let out = match l with
            | Text s | Space s -> s
            | Open t -> sprintf "<%s>" (string_of_tag t)
            | Close t -> sprintf "</%s>" (string_of_tag t)
            | EOF -> assert false
  in
  printf "%s" out;
  column := !column + String.length out;
;;

type attribute = {
 b : bool;
 em : bool;
 i : bool;
 s : bool;
 tt : bool;
 u : int; 
 size : int;
 color : color;
};;

let plain = {
  b = false;
  i = false;
  em = false;
  s = false;
  tt = false;
  u = 0; 
  size = -1;
  color = Initial;
};;

let random_tag () =
  match Random.int 25 with
  | (0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9) as s -> Size s
  | 10 -> B
  | 11 -> EM
  | 12 -> I
  | 13 -> S
  | 14 -> TT
  | 15 -> U
  | 16 -> Color Red
  | 17 -> Color Green
  | 18 -> Color Blue
  | 19 -> Color Cyan
  | 20 -> Color Magenta
  | 21 -> Color Yellow
  | 22 -> Color Black
  | 23 -> Color White
  | 24 -> PL
  | _ -> assert false
;;

let is_size = function Size _ -> true | _ -> false;;
let is_color = function Color _ -> true | _ -> false;;

let pushsame = ref false;;
let rndshape = ref false;;

let rec new_tag ctx top =
  let t = random_tag () in
  match t with
  | B -> if ctx.b then new_tag ctx top else (t, {ctx with b = true})
  | EM ->
      if ctx.s || top = EM
      then new_tag ctx top
      else (t, {ctx with em = not ctx.em})
  | I -> if ctx.i then new_tag ctx top else (t, {ctx with i = true})
  | S -> if ctx.s then new_tag ctx top else (t, {ctx with s = true})
  | TT -> if ctx.tt then new_tag ctx top else (t, {ctx with tt = true})
  | U -> if ctx.u >= 3 then new_tag ctx top else (t, {ctx with u = ctx.u + 1})
  | Size s ->
      if ctx.size = s || is_size top && not !pushsame
      then new_tag ctx top
      else (t, {ctx with size = s})
  | Color c ->
      if ctx.color = c || is_color top && not !pushsame
      then new_tag ctx top
      else (t, {ctx with color = c})
  | PL ->
      if not ctx.b && not ctx.em && not ctx.i && not ctx.s && not ctx.tt
         && ctx.u = 0
      then new_tag ctx top
      else (t, {plain with size = ctx.size; color = ctx.color})
;;  

(* if rndshape is false, then print default;
   else randomly choose between "." and nothing.
*)
let print_rnd_text default =
  if not !rndshape then print_lexeme (Text default)
  else if Random.int 2 = 0 then print_lexeme (Text ".")
;;

let rec recurse ctx top w d =
  let nd = d - 1 in
  let (t, nctx) = new_tag ctx top in
  print_lexeme (Open t);
  if d > 0 then begin
    let this_w = if w = 0 then 1 + Random.int 4 else w in
    print_rnd_text "";
    recurse nctx t w nd;
    for i = 2 to this_w do
      print_rnd_text ".";
      recurse nctx (if i = this_w then t else B) w nd;
    done;
    print_rnd_text "";
  end else begin
    print_lexeme (Text ".");
  end;
  print_lexeme (Close t);
;;

let usage = "usage: fractal [options]\n";;

let depth = ref 3;;
let seed = ref 1;;
let width = ref 0;;

let options = [
  "-depth", Arg.Int ((:=) depth), "depth of the tree (default: 3)";
  "-rndshape", Arg.Set rndshape, "use random shape";
  "-pushsame", Arg.Set pushsame, "can push color on color, size on size";
  "-seed", Arg.Int ((:=) seed), "PRNG seed (default 1)";
  "-width", Arg.Int ((:=) width), "degree of each node (default: random 1-5)";
];;

let anon x = raise (Arg.Bad ("unknown option: " ^ x));;

let main () =
  Arg.parse options anon usage;
  Random.init !seed;
  recurse plain PL !width !depth;
  printf "\n";
;;

main ();;
