open Printf;;

type color =
  | Red | Green | Blue | Cyan | Magenta | Yellow | Black | White | Initial
;;

let string_of_color = function
 | Red -> "r"
 | Green -> "g"
 | Blue -> "b"
 | Cyan -> "c"
 | Magenta -> "m"
 | Yellow -> "y"
 | Black -> "k"
 | White -> "w"
 | Initial -> "i"
;;

type attribute = {
 u : int; 
 s : bool;
 em : bool;
 i : bool;
 b : bool;
 tt : bool;
 size : int;
 color : color;
};;

type tag_type =
 | U
 | S
 | EM
 | I
 | B
 | TT
 | PL
 | Size of int
 | Color of color
;;

let string_of_tag = function
 | U -> "U"
 | S -> "S"
 | EM -> "EM"
 | I -> "I"
 | B -> "B"
 | TT -> "TT"
 | PL -> "PL"
 | Size int -> string_of_int int
 | Color color -> string_of_color color
;;

type lexem =
   | Text of string
   | Space of string
   | Open of tag_type
   | Close of tag_type
   | EOF
;;

type decorated_char = char * attribute;;

let error s = failwith s;;

let plain = {
  u = 0; 
  s = false;
  em = false;
  i = false;
  b = false;
  tt = false;
  size = -1;
  color = Initial;
};;

let rec extract_nth l n =
  match (l, n) with
    | h::t, 0 -> (h, t)
    | h::t, n -> let (e, l1) = extract_nth t (n-1) in (e, h::l1)
    | _,    _ -> assert false
;;

let rec shuffle = function
  | [] -> []
  | [_] as l -> l
  | l ->
     let n = Random.int (List.length l) in
     let (e, l1) = extract_nth l n in
     e :: (shuffle l1)
;;

let mktags ctx =
  let l = if ctx.b then [B] else [] in
  let l = if ctx.em then EM::l else l in
  let l = if ctx.i then I::l else l in
  let l = if ctx.s then S::EM::l else l in
  let l = if ctx.tt then TT::l else l in
  let l = match ctx.u with
          | 0 -> l
          | 1 -> U::l
          | 2 -> U::U::l
          | _ -> U::U::U::l
  in
  let l = if ctx.size <> -1 then (Size ctx.size)::l else l in
  let l = if ctx.color <> Initial then (Color ctx.color)::l else l in
  shuffle l
;;

let rec tok = parser
 | [< ''<'; t = lex_tag >] -> t
 | [< ' (' ' | '\009' | '\010' | '\013' as c); strm >]
     -> let b = Buffer.create 10 in
        Buffer.add_char b c;
        lex_spaces b strm
 | [< 'c; strm >]
     -> let b = Buffer.create 10 in
        Buffer.add_char b c;
        lex_text b strm
 | [< >] -> EOF

and lex_tag_type = parser
 | [< ''U'; ''>' >] -> U
 | [< ''S'; ''>' >] -> S
 | [< ''E'; ''M'; ''>' >] -> EM
 | [< ''I'; ''>' >] -> I
 | [< ''B'; ''>' >] -> B
 | [< ''T'; ''T'; ''>' >] -> TT
 | [< ''P'; ''L'; ''>' >] -> PL
 | [< ''r'; ''>' >] -> Color Red
 | [< ''g'; ''>' >] -> Color Green
 | [< ''b'; ''>' >] -> Color Blue
 | [< ''c'; ''>' >] -> Color Cyan
 | [< ''m'; ''>' >] -> Color Magenta
 | [< ''y'; ''>' >] -> Color Yellow
 | [< ''k'; ''>' >] -> Color Black
 | [< ''w'; ''>' >] -> Color White
 | [< ''0'..'9' as c; ''>' >] -> Size (Char.code c - Char.code '0')

and lex_tag = parser
 | [< ''/'; s >] -> Close (lex_tag_type s)
 | [< ty = lex_tag_type >] -> Open ty

and lex_spaces b = parser
  | [< ' (' ' | '\009' | '\010' | '\013' as c); strm >]
      -> Buffer.add_char b c;
         lex_spaces b strm
  | [< >] -> Space (Buffer.contents b)

and lex_text b = parser
  | [< ' ('\033'..'\059' | '\061'..'\126' as c); strm >]
      -> Buffer.add_char b c;
         lex_text b strm
  | [< >] -> Text (Buffer.contents b)
;;

let rec print_tags prefix = function
  | [] -> ()
  | h::t -> printf "<%s%s>" prefix (string_of_tag h);
            print_tags prefix t;
;;

let print_open_tags l = print_tags "" l;;

let print_close_tags l = print_tags "/" (List.rev l);;
 
let rec pessim ctx stack strm =
  match tok strm with
  | EOF -> if stack <> [] then error "unclosed tag"
  | Text s ->
      let tags = mktags ctx in
      print_open_tags tags;
      print_string s;
      print_close_tags tags;
      pessim ctx stack strm;
  | Space s ->
      let tags = mktags ctx in
      print_open_tags tags;
      print_string s;
      print_close_tags tags;
      pessim ctx stack strm;
  | Open t ->
      let newctx = match t with
        | B -> {ctx with b = true}
        | EM -> {ctx with em = not ctx.em}
        | I -> {ctx with i = true}
        | PL -> {plain with size = ctx.size; color = ctx.color}
        | S -> {ctx with s = true}
        | TT -> {ctx with tt = true}
        | U -> {ctx with u = if ctx.u >= 3 then 3 else 1 + ctx.u}
        | Size n -> {ctx with size = n}
        | Color c -> {ctx with color = c}
      in
      let newctx = if newctx = ctx then ctx else newctx in
      pessim newctx (ctx::stack) strm;
   | Close t ->
      begin match stack with
        | [] -> error "unmatched closing tag"
        | h::t -> pessim h t strm;
      end;
;;

let main () =
  pessim plain [] (Stream.of_channel stdin);
  flush stdout;
;;

main ();;
