open Printf;;

type color =
  | Red | Green | Blue | Cyan | Magenta | Yellow | Black | White | Initial
;;

let string_of_color = function
 | Red -> "r"
 | Green -> "g"
 | Blue -> "b"
 | Cyan -> "c"
 | Magenta -> "m"
 | Yellow -> "y"
 | Black -> "k"
 | White -> "w"
 | Initial -> assert false
;;

type tag_type =
 | B
 | EM
 | I
 | PL
 | S
 | TT
 | U
 | Size of int
 | Color of color
;;

let string_of_tag = function
 | U -> "U"
 | S -> "S"
 | EM -> "EM"
 | I -> "I"
 | B -> "B"
 | TT -> "TT"
 | PL -> "PL"
 | Size (-1) -> assert false
 | Size int -> string_of_int int
 | Color color -> string_of_color color
;;

type lexem =
   | Text of string
   | Space of string
   | Open of tag_type
   | Close of tag_type
   | EOF
;;

let rec tok = parser
 | [< ''>'; t = lex_tag; ''<' >] -> t
 | [< ' (' ' | '\009' | '\010' | '\013' as c); strm >]
     -> let b = Buffer.create 10 in
        Buffer.add_char b c;
        lex_spaces b strm
 | [< 'c; strm >]
     -> let b = Buffer.create 10 in
        Buffer.add_char b c;
        lex_text b strm
 | [< >] -> EOF

and lex_tag_type = parser
 | [< ''U' >] -> U
 | [< ''S' >] -> S
 | [< ''M'; ''E'; >] -> EM
 | [< ''I'; >] -> I
 | [< ''B'; >] -> B
 | [< ''T'; ''T' >] -> TT
 | [< ''L'; ''P' >] -> PL
 | [< ''r' >] -> Color Red
 | [< ''g' >] -> Color Green
 | [< ''b' >] -> Color Blue
 | [< ''c' >] -> Color Cyan
 | [< ''m' >] -> Color Magenta
 | [< ''y' >] -> Color Yellow
 | [< ''k' >] -> Color Black
 | [< ''w' >] -> Color White
 | [< ''0'..'9' as c >] -> Size (Char.code c - Char.code '0')

and lex_tag = parser
 | [< ty = lex_tag_type; s >] -> lex_openclose ty s

and lex_openclose ty = parser
 | [< ''/' >] -> Open ty
 | [< >] -> Close ty

and lex_spaces b = parser
  | [< ' (' ' | '\009' | '\010' | '\013' as c); strm >]
      -> Buffer.add_char b c;
         lex_spaces b strm
  | [< >] -> Space (Buffer.contents b)

and lex_text b = parser
  | [< 'c when c <> '<' && c <> '>'; strm >]
      -> Buffer.add_char b c;
         lex_text b strm
  | [< >] -> Text (Buffer.contents b)
;;

let print_lexeme = function
  | Text s | Space s -> printf "%s" s;
  | Open t -> printf "<%s>" (string_of_tag t)
  | Close t -> printf "</%s>" (string_of_tag t)
  | EOF -> assert false
;;

let read_stream s =
  let buf = String.create 8192 in
  let b = Buffer.create 1000 in
  let rec loop () =
    let l = input s buf 0 8192 in
    if l > 0 then begin
      Buffer.add_substring b buf 0 l;
      loop ();
    end
  in
  loop ();
  Buffer.contents b
;;

(* Reverse the string IN PLACE *)
let reverse s =
  let l = String.length s - 1 in
  for i = 0 to l / 2 do
    let c1 = s.[i] in
    let c2 = s.[l-i] in
    s.[i] <- c2;
    s.[l-i] <- c1;
  done;
;;

let main () =
  let doc = read_stream stdin in
  reverse doc;
  let str = Stream.of_string doc in
  let rec loop () =
    match tok str with
    | EOF -> ()
    | t -> print_lexeme t; loop ();
  in
  loop ();
;;

main ();;
