open Sml;;

let string_of_color = function
 | Red -> "r"
 | Green -> "g"
 | Blue -> "b"
 | Cyan -> "c"
 | Magenta -> "m"
 | Yellow -> "y"
 | Black -> "k"
 | White -> "w"
 | Initial -> "i"
;;

let string_of_tag = function
 | U -> "U"
 | S -> "S"
 | EM -> "EM"
 | I -> "I"
 | B -> "B"
 | TT -> "TT"
 | PL -> "PL"
 | SZ int -> string_of_int int
 | Color color -> string_of_color color
;;

let error s = failwith s;;

let plain = {
  u = 0; 
  s = false;
  em = false;
  i = false;
  b = false;
  tt = false;
  size = -1;
  color = Initial;
 };;

let rec sem prev ctx stack accu l =

  let add_prev prev accu =
   match prev with
   | None -> accu
   | Some (s, ctx as tok) -> tok :: accu in

  match l with
  | [] ->
     if stack = [] then add_prev prev accu
     else error ("Unmatched open tag in output: " ^
                  string_of_tag (fst (List.hd stack)))
  | tok :: rest ->
      match tok with
      | Text c ->
          begin match prev with
          | Some (c0, pctx) when pctx = ctx ->
             sem (Some (c0 ^ c, ctx)) ctx stack accu rest
          | _ ->
             sem (Some (c, ctx)) ctx stack (add_prev prev accu) rest end
      | Space c ->
          let spcctx =
             {ctx with
              s = false;
              em = false;
              i = false;
              b = false;
              color = if ctx.u = 0 then Black else ctx.color}
          in
          if ctx.tt then
            sem (Some (String.make 1 c, spcctx)) ctx stack
                (add_prev prev accu) rest
          else
          let cur = Some (" ", spcctx) in
          if cur = prev then sem prev ctx stack accu rest else
          sem cur ctx stack (add_prev prev accu) rest
       | Open tag ->
           let nctx =
            match tag with
            | U -> if ctx.u < 3 then {ctx with u = ctx.u + 1} else ctx
            | S -> if ctx.s && not ctx.em then ctx
                   else {ctx with s = true; em = false}
            | EM -> if not ctx.s then {ctx with em = not ctx.em} else ctx
            | I -> if ctx.i then ctx else {ctx with i = true}
            | B -> if ctx.b then ctx else {ctx with b = true}
            | TT -> if ctx.tt then ctx else {ctx with tt = true}
            | PL ->
                let nc = {plain with size = ctx.size; color = ctx.color} in
                if nc = ctx then ctx else nc
            | SZ i -> if ctx.size = i then ctx else {ctx with size = i}
            | Color c -> if ctx.color = c then ctx else {ctx with color = c} in
           sem prev nctx ((tag, ctx) :: stack) accu rest
        | Close ctag ->
            match stack with
            | [] -> error "Empty stack in semantics analysis"
            | (otag, ctx) :: stack ->
                if otag <> ctag
                then error ("Unmatched close tag in output : " ^
                            string_of_tag ctag)
                else sem prev ctx stack accu rest;;

let init_tok = None;;

let semant l =
 sem init_tok plain [] [] l;;

let rec lsem prev ctx stack l =

  let add_prev prev accu =
   match prev with
   | None -> Lazy.force accu
   | Some (s, ctx as tok) -> Lcons (tok, accu) in

  match l with
  | Lnil ->
     if stack = [] then add_prev prev (lazy Lnil) else
     error ("Unmatched open tag in output: " ^
             string_of_tag (fst (List.hd stack)))
  | Lcons (tok, rest) ->
      match tok with
      | Text c ->
          begin match prev with
          | Some (c0, pctx) when pctx = ctx ->
             lsem (Some (c0 ^ c, ctx)) ctx stack (Lazy.force rest)
          | _ ->
             let nprev = Some (c, ctx) in
             add_prev prev (lazy (lsem nprev ctx stack (Lazy.force rest))) end
      | Space c ->
          let spcctx =
             {ctx with
              s = false;
              em = false;
              i = false;
              b = false;
              color = if ctx.u = 0 then Black else ctx.color}
          in
          if ctx.tt then
           let cur = Some (String.make 1 c, spcctx) in
           add_prev prev (lazy (lsem cur ctx stack (Lazy.force rest)))
          else
          let cur = Some (" ", spcctx) in
          if cur = prev
          then lsem prev ctx stack (Lazy.force rest)
          else add_prev prev (lazy (lsem cur ctx stack (Lazy.force rest)))
        | Open tag ->
           let nctx =
            match tag with
            | U -> if ctx.u < 3 then {ctx with u = ctx.u + 1} else ctx
            | S -> {ctx with s = true; em = false}
            | EM -> if not ctx.s then {ctx with em = not ctx.em} else ctx
            | I -> if ctx.i then ctx else {ctx with i = true}
            | B -> if ctx.b then ctx else {ctx with b = true}
            | TT -> if ctx.tt then ctx else {ctx with tt = true}
            | PL -> let nc = {plain with size = ctx.size; color = ctx.color} in
                    if nc = ctx then ctx else nc
            | SZ i -> if ctx.size = i then ctx else {ctx with size = i}
            | Color c -> if ctx.color = c then ctx else {ctx with color = c} in
           lsem prev nctx ((tag, ctx) :: stack) (Lazy.force rest)
        | Close ctag ->
            match stack with
            | [] -> error "Empty stack in semantics analysis"
            | (otag, ctx) :: stack ->
                if otag <> ctag
                then error ("Unmatched close tag in output : " ^
                            string_of_tag ctag)
                else lsem prev ctx stack (Lazy.force rest);;

let lsemant l =
 lsem init_tok plain [] l;;

let rec lequal l1 l2 =
 match l1, l2 with
 | Lnil, Lnil -> true
 | Lcons (h1, r1), Lcons (h2, r2) ->
    h1 = h2 && lequal (Lazy.force r1) (Lazy.force r2)
 | _, _ -> false;;

