open Printf;;

type color =
  | Red | Green | Blue | Cyan | Magenta | Yellow | Black | White | Initial
;;

let string_of_color = function
 | Red -> "r"
 | Green -> "g"
 | Blue -> "b"
 | Cyan -> "c"
 | Magenta -> "m"
 | Yellow -> "y"
 | Black -> "k"
 | White -> "w"
 | Initial -> assert false
;;

type attribute = {
 b : bool;
 em : bool;
 i : bool;
 s : bool;
 tt : bool;
 u : int; 
 size : int;
 color : color;
};;

type space_context = {
  sp_tt : bool;
  sp_u : int;
  sp_size : int;
  sp_color : color;
};;

type tag_type =
 | B
 | EM
 | I
 | PL
 | S
 | TT
 | U
 | Size of int
 | Color of color
;;

let string_of_tag = function
 | U -> "U"
 | S -> "S"
 | EM -> "EM"
 | I -> "I"
 | B -> "B"
 | TT -> "TT"
 | PL -> "PL"
 | Size (-1) -> assert false
 | Size int -> string_of_int int
 | Color color -> string_of_color color
;;

type lexem =
   | Text of string
   | Space of string
   | Open of tag_type
   | Close of tag_type
   | EOF
;;

let error s = failwith s;;

let plain = {
  b = false;
  i = false;
  em = false;
  s = false;
  tt = false;
  u = 0; 
  size = -1;
  color = Initial;
};;

let plain_sp = {
  sp_u = 0;
  sp_tt = false;
  sp_size = -1;
  sp_color = Initial;
};;

let rec tok = parser
 | [< ''<'; t = lex_tag >] -> t
 | [< ' (' ' | '\009' | '\010' | '\013' as c); strm >]
     -> let b = Buffer.create 10 in
        Buffer.add_char b c;
        lex_spaces b strm
 | [< 'c; strm >]
     -> let b = Buffer.create 10 in
        Buffer.add_char b c;
        lex_text b strm
 | [< >] -> EOF

and lex_tag_type = parser
 | [< ''U'; ''>' >] -> U
 | [< ''S'; ''>' >] -> S
 | [< ''E'; ''M'; ''>' >] -> EM
 | [< ''I'; ''>' >] -> I
 | [< ''B'; ''>' >] -> B
 | [< ''T'; ''T'; ''>' >] -> TT
 | [< ''P'; ''L'; ''>' >] -> PL
 | [< ''r'; ''>' >] -> Color Red
 | [< ''g'; ''>' >] -> Color Green
 | [< ''b'; ''>' >] -> Color Blue
 | [< ''c'; ''>' >] -> Color Cyan
 | [< ''m'; ''>' >] -> Color Magenta
 | [< ''y'; ''>' >] -> Color Yellow
 | [< ''k'; ''>' >] -> Color Black
 | [< ''w'; ''>' >] -> Color White
 | [< ''0'..'9' as c; ''>' >] -> Size (Char.code c - Char.code '0')

and lex_tag = parser
 | [< ''/'; s >] -> Close (lex_tag_type s)
 | [< ty = lex_tag_type >] -> Open ty

and lex_spaces b = parser
  | [< ' (' ' | '\009' | '\010' | '\013' as c); strm >]
      -> Buffer.add_char b c;
         lex_spaces b strm
  | [< >] -> Space (Buffer.contents b)

and lex_text b = parser
  | [< ' ('\033'..'\059' | '\061'..'\126' as c); strm >]
      -> Buffer.add_char b c;
         lex_text b strm
  | [< >] -> Text (Buffer.contents b)
;;

let column = ref 0;;

let print_open_tag t =
  printf "<%s>" (string_of_tag t);
  column := !column + 2 + String.length (string_of_tag t);
;;
let print_close_tag t =
  printf "</%s>" (string_of_tag t);
  column := !column + 3 + String.length (string_of_tag t);
;;

let rec is_compatible stack ctx =
  match stack with
  | [] -> true
  | B :: t -> ctx.b && is_compatible t ctx
  | EM :: t -> is_compatible t ctx
  | I :: t -> ctx.i && is_compatible t ctx
  | PL :: t -> assert false
  | S :: t -> ctx.s && is_compatible t ctx
  | TT :: t -> ctx.tt && is_compatible t ctx
  | U :: t -> ctx.u > 0 && is_compatible t {ctx with u = ctx.u - 1}
  | (Size n) :: t -> ctx.size <> -1 && is_compatible t ctx
  | (Color c) :: t -> ctx.color <> Initial && is_compatible t ctx
;;

let rec is_compatible_sp u stack ctx =
  match stack with
  | [] -> true
  | (B | EM | I | PL | S) :: t -> is_compatible_sp u t ctx
  | TT :: t -> ctx.sp_tt && is_compatible_sp u t ctx
  | U :: t ->
      ctx.sp_u > 0 && is_compatible_sp u t {ctx with sp_u = ctx.sp_u - 1}
  | (Size n) :: t -> ctx.sp_size <> -1 && is_compatible_sp u t ctx
  | (Color c) :: t when u = 0 -> is_compatible_sp u t ctx
  | (Color c) :: t -> ctx.sp_color <> Initial && is_compatible_sp u t ctx
;;

let list_count e l =
  let rec count = function
    | [] -> 0
    | h::t when h = e -> 1 + count t
    | h::t -> count t
  in
  count l
;;

let rec first_size = function
  | [] -> -1
  | Size s :: t -> s
  | _ :: t -> first_size t
;;

let rec first_color = function
  | [] -> Initial
  | Color c :: t -> c
  | _ :: t -> first_color t
;;

let em_at_head = function
  | EM :: _ -> true
  | _ -> false
;;

let rec list_index e l n =
  match l with
  | [] -> n
  | h::t when e = h -> n
  | _::t -> list_index e t (n+1)
;;

(* t1 is smaller if it occurs before t2 in intags *)
let cmptags intags =
  let ib = ref 0 and iem = ref 0 and ii = ref 0 and is = ref 0
  and itt = ref 0 and iu = ref 0 and isz = Array.make 10 0
  and ired = ref 0 and igreen = ref 0 and iblue = ref 0
  and icyan = ref 0 and imagenta = ref 0 and iyellow = ref 0
  and iblack = ref 0 and iwhite = ref 0
  in
  let rec loop i l =
    match l with
      | [] -> ()
      | h::t ->
          begin match h with
            | B -> ib := i
            | EM -> iem := i
            | I -> ii := i
            | PL -> ()
            | S -> is := i
            | TT -> itt := i
            | U -> iu := i
            | Size (-1) -> ()
            | Size s -> isz.(s) <- i
            | Color Red -> ired := i
            | Color Green -> igreen := i
            | Color Blue -> iblue := i
            | Color Cyan -> icyan := i
            | Color Magenta -> imagenta := i
            | Color Yellow -> iyellow := i
            | Color Black -> iblack := i
            | Color White -> iwhite := i
            | Color Initial -> ()
          end;
          loop (i+1) t;
  in
  let idx = function
    | B -> !ib
    | EM -> !iem
    | I -> !ii
    | PL -> assert false
    | S -> !is
    | TT -> !itt
    | U -> !iu
    | Size (-1) -> assert false
    | Size s -> isz.(s)
    | Color Red -> !ired
    | Color Green -> !igreen
    | Color Blue -> !iblue
    | Color Cyan -> !icyan
    | Color Magenta -> !imagenta
    | Color Yellow -> !iyellow
    | Color Black -> !iblack
    | Color White -> !iwhite
    | Color Initial -> assert false
  in
  loop 0 intags;
  fun t1 t2 -> idx t1 - idx t2
;;

let rec list_make e n =
  if n <= 0 then [] else e :: (list_make e (n-1))
;;

let rec stack_to_ctx = function
  | [] -> plain
  | B::t -> {(stack_to_ctx t) with b = true}
  | EM::t -> let c = stack_to_ctx t in {c with em = not c.em}
  | I::t -> {(stack_to_ctx t) with i = true}
  | PL::t -> assert false
  | S::t -> {(stack_to_ctx t) with s = true}
  | TT::t -> {(stack_to_ctx t) with tt = true}
  | U::t -> let c = stack_to_ctx t in {c with u = c.u + 1}
  | Size s :: t -> {(stack_to_ctx t) with size = s}
  | Color co :: t -> {(stack_to_ctx t) with color = co}
;;

let add_missing stack ctx intags =
  let sctx = stack_to_ctx stack in
  let newtags = List.concat [
        (list_make U (ctx.u - sctx.u));
        (if ctx.tt && not sctx.tt then [TT] else []);
        (if ctx.s && not sctx.s then [S] else []);
        (if ctx.i && not sctx.i then [I] else []);
        (if ctx.b && not sctx.b then [B] else []);
        (if ctx.s = false && ctx.em <> sctx.em && not (em_at_head stack)
           then [EM] else []);
        (if ctx.color <> Initial && sctx.color <> ctx.color
           then [Color ctx.color] else []);
        (if ctx.size <> -1 && sctx.size <> ctx.size
           then [Size ctx.size] else []);
      ]
  in
  let stk = if ctx.s = false && ctx.em <> sctx.em && em_at_head stack
            then List.tl stack
            else stack
  in (List.sort (cmptags intags) newtags) @ stk
;;

let add_missing_sp stack ctx intags =
  let sctx = stack_to_ctx stack in
  let newtags = List.concat [
        (list_make U (ctx.sp_u - sctx.u));
        (if ctx.sp_tt && not sctx.tt then [TT] else []);
        (if ctx.sp_color <> Initial && ctx.sp_u > 0
            && sctx.color <> ctx.sp_color
           then [Color ctx.sp_color] else []);
        (if ctx.sp_size <> -1 && sctx.size <> ctx.sp_size
           then [Size ctx.sp_size] else []);
      ]
    in (List.sort (cmptags intags) newtags) @ stack
;;

let rec update_stack stack ctx intags =
  if is_compatible stack ctx
  then add_missing stack ctx intags
  else update_stack (List.tl stack) ctx intags
;;

let rec update_stack_sp stack spctx intags =
  if is_compatible_sp spctx.sp_u stack spctx
  then add_missing_sp stack spctx intags
  else update_stack_sp (List.tl stack) spctx intags
;;

let rec print_close_tags_diff stack suf =
  if stack = suf then () else begin
    assert (stack <> []);
    print_close_tag (List.hd stack);
    print_close_tags_diff (List.tl stack) suf;
  end;
;;

let rec print_open_tags_diff stack suf =
  if stack = suf then () else begin
    assert (stack <> []);
    print_open_tags_diff (List.tl stack) suf;
    print_open_tag (List.hd stack);
  end;
;;

let rec common_suffix l1 l2 =
  if List.length l1 > List.length l2 then common_suffix (List.tl l1) l2
  else if List.length l1 < List.length l2 then common_suffix l1 (List.tl l2)
  else if l1 = l2 then l1
  else common_suffix (List.tl l1) (List.tl l2)
;;

let print_diff_tags ostack nstack =
  let suf = common_suffix ostack nstack in
  print_close_tags_diff ostack suf;
  print_open_tags_diff nstack suf;
;;

let same_spctx c1 c2 =
  c1.sp_tt = c2.sp_tt
  && c1.sp_size = c2.sp_size
  && c1.sp_u = c2.sp_u
  && (c1.sp_u = 0 || c1.sp_color = c2.sp_color)
;;

let print s =
  printf "%s" s;
  for i = 0 to String.length s - 1 do
    match s.[i] with
    | '\n' -> column := 0;
    | '\t' -> column := !column + 8;
    | _ -> incr column;
  done;
;;

let rec optim prevstr prevspctx outstack ctx instack intags strm =
  match tok strm with
  | EOF -> print_diff_tags outstack [];
  | Text s ->
      let newstack = update_stack outstack ctx intags in
      print_diff_tags outstack newstack;
      print s;
      optim s prevspctx newstack ctx instack intags strm;
  | Space s ->
      let newspctx = {sp_tt = ctx.tt; sp_size = ctx.size; sp_u = ctx.u;
                      sp_color = ctx.color}
      in
      if ctx.tt then begin
        let newstack = update_stack_sp outstack newspctx intags in
        print_diff_tags outstack newstack;
        print s;
        optim s newspctx newstack ctx instack intags strm;
      end else if prevstr = " " && same_spctx newspctx prevspctx then begin
        optim prevstr prevspctx outstack ctx instack intags strm;
      end else begin
        let newstack = update_stack_sp outstack newspctx intags in
        print_diff_tags outstack newstack;
        if !column > 50 then print "\n" else print "\n";
        optim " " newspctx newstack ctx instack intags strm;
      end
  | Open t ->
      let newctx = match t with
        | B -> {ctx with b = true}
        | EM -> {ctx with em = if ctx.s then ctx.em else not ctx.em}
        | I -> {ctx with i = true}
        | PL -> {plain with size = ctx.size; color = ctx.color}
        | S -> {ctx with s = true}
        | TT -> {ctx with tt = true}
        | U -> {ctx with u = if ctx.u >= 3 then 3 else 1 + ctx.u}
        | Size n -> {ctx with size = n}
        | Color c -> {ctx with color = c}
      in
      optim prevstr prevspctx outstack newctx (ctx::instack) (t::intags) strm;
   | Close t ->
      begin match instack with
        | [] -> error "unmatched closing tag"
        | h::t -> optim prevstr prevspctx outstack h t (List.tl intags) strm;
      end;
;;

let main () =
  optim "" plain_sp [] plain [] [] (Stream.of_channel stdin);
  flush stdout;
;;

main ();;
