open Printf;;

type color =
  | Red | Green | Blue | Cyan | Magenta | Yellow | Black | White | Initial
;;

let string_of_color = function
 | Red -> "r"
 | Green -> "g"
 | Blue -> "b"
 | Cyan -> "c"
 | Magenta -> "m"
 | Yellow -> "y"
 | Black -> "k"
 | White -> "w"
 | Initial -> assert false
;;

type attribute = {
 b : bool;
 em : bool;
 i : bool;
 s : bool;
 tt : bool;
 u : int; 
 size : int;
 color : color;
};;

type space_context = {
  sp_tt : bool;
  sp_u : int;
  sp_size : int;
  sp_color : color;
};;

type tag_type =
 | B
 | EM
 | I
 | PL
 | S
 | TT
 | U
 | Size of int
 | Color of color
;;

let string_of_tag = function
 | U -> "U"
 | S -> "S"
 | EM -> "EM"
 | I -> "I"
 | B -> "B"
 | TT -> "TT"
 | PL -> "PL"
 | Size (-1) -> assert false
 | Size int -> string_of_int int
 | Color color -> string_of_color color
;;

type lexem =
   | Text of string
   | Space of string
   | Open of tag_type
   | Close of tag_type
   | EOF
;;

(********************************************************************)

type opens = {
  mutable ordered : tag_type list;
  mutable free : tag_type list;
};;

type set =
  | String of string
  | Opens of opens
  | Closes of tag_type list
;;

type fifo = {
  mutable front : set list;
  mutable back : set list;
}

let fifo = { front = []; back = [] };;

let fifo_transfer () =
  fifo.front <- List.rev_append fifo.back fifo.front;
  fifo.back <- [];
;;
let fifo_put x = fifo.back <- x :: fifo.back;;
let fifo_head () =
  if fifo.front = [] then fifo_transfer ();
  match fifo.front with
  | h::_ -> Some h
  | []   -> None
;;
let fifo_get () =
  if fifo.front = [] then fifo_transfer ();
  match fifo.front with
  | h::t -> fifo.front <- t; Some h
  | []   -> None
;;

(********************************************************************)

let error s = failwith s;;

let plain = {
  b = false;
  i = false;
  em = false;
  s = false;
  tt = false;
  u = 0; 
  size = -1;
  color = Initial;
};;

let plain_sp = {
  sp_u = 0;
  sp_tt = false;
  sp_size = -1;
  sp_color = Initial;
};;

let rec tok = parser
 | [< ''<'; t = lex_tag >] -> t
 | [< ' (' ' | '\009' | '\010' | '\013' as c); strm >]
     -> let b = Buffer.create 10 in
        Buffer.add_char b c;
        lex_spaces b strm
 | [< 'c; strm >]
     -> let b = Buffer.create 10 in
        Buffer.add_char b c;
        lex_text b strm
 | [< >] -> EOF

and lex_tag_type = parser
 | [< ''U'; ''>' >] -> U
 | [< ''S'; ''>' >] -> S
 | [< ''E'; ''M'; ''>' >] -> EM
 | [< ''I'; ''>' >] -> I
 | [< ''B'; ''>' >] -> B
 | [< ''T'; ''T'; ''>' >] -> TT
 | [< ''P'; ''L'; ''>' >] -> PL
 | [< ''r'; ''>' >] -> Color Red
 | [< ''g'; ''>' >] -> Color Green
 | [< ''b'; ''>' >] -> Color Blue
 | [< ''c'; ''>' >] -> Color Cyan
 | [< ''m'; ''>' >] -> Color Magenta
 | [< ''y'; ''>' >] -> Color Yellow
 | [< ''k'; ''>' >] -> Color Black
 | [< ''w'; ''>' >] -> Color White
 | [< ''0'..'9' as c; ''>' >] -> Size (Char.code c - Char.code '0')

and lex_tag = parser
 | [< ''/'; s >] -> Close (lex_tag_type s)
 | [< ty = lex_tag_type >] -> Open ty

and lex_spaces b = parser
  | [< ' (' ' | '\009' | '\010' | '\013' as c); strm >]
      -> Buffer.add_char b c;
         lex_spaces b strm
  | [< >] -> Space (Buffer.contents b)

and lex_text b = parser
  | [< ' ('\033'..'\059' | '\061'..'\126' as c); strm >]
      -> Buffer.add_char b c;
         lex_text b strm
  | [< >] -> Text (Buffer.contents b)
;;

let list_count e l =
  let rec count = function
    | [] -> 0
    | h::t when h = e -> 1 + count t
    | h::t -> count t
  in
  count l
;;

let rec list_first_size = function
  | [] -> -1
  | Size s :: t -> s
  | _ :: t -> list_first_size t
;;

let rec list_first_color = function
  | [] -> Initial
  | Color c :: t -> c
  | _ :: t -> list_first_color t
;;

let rec stack_count e l st =
  list_count e l
  + ( match st with
      | [] -> 0
      | { ordered = _; free = l2 } :: st2 -> stack_count e l2 st2
    )
;;

let stack_mem e l st = stack_count e l st > 0;;

let rec stack_first_size l st =
  let lf = list_first_size l in
  if lf <> -1 then lf else begin
    match st with
    | [] -> -1
    | { ordered = _; free = l2 } :: st2 -> stack_first_size l2 st2
  end
;;

let rec stack_first_color l st =
  let lf = list_first_color l in
  if lf <> Initial then lf else begin
    match st with
    | [] -> Initial
    | { ordered = _; free = l2 } :: st2 -> stack_first_color l2 st2
  end
;;

let (==>) a b = b || not a;;

let rec add_tags c = function
  | [] -> c
  | B::t -> add_tags {c with b = true} t
  | EM::t -> add_tags {c with em = true} t
  | I::t -> add_tags {c with i = true} t
  | PL::t -> assert false
  | S::t -> add_tags {c with s = true} t
  | TT::t -> add_tags {c with tt = true} t
  | U::t -> add_tags {c with u = c.u + 1} t
  | Size s :: t -> add_tags {c with size = s} t
  | Color co :: t -> add_tags {c with color = co} t
;;

let rec stack_to_ctx top = function
  | [] -> add_tags plain top
  | {free = l} :: t ->
      let c2 = stack_to_ctx l t in
      add_tags c2 top
;;

let rec is_compatible top stack ctx =
  let sctx = stack_to_ctx top stack in
  (sctx.b ==> ctx.b)
  && (sctx.i ==> ctx.i)
  && (sctx.s ==> ctx.s)
  && (sctx.tt ==> ctx.tt)
  && (sctx.u <= ctx.u)
  && ((ctx.size = -1) ==> (sctx.size = -1))
  && ((ctx.color = Initial) ==> (sctx.color = Initial))
;;

let rec is_compatible_sp top stack ctx =
  let sctx = stack_to_ctx top stack in
  (sctx.tt ==> ctx.sp_tt)
  && (sctx.u <= ctx.sp_u)
  && ((ctx.sp_size = -1) ==> (sctx.size = -1))
  && ((ctx.sp_u > 0 && ctx.sp_color = Initial) ==> (sctx.color = Initial))
;;

let column = ref 0;;

let print s =
  printf "%s" s;
  for i = 0 to String.length s - 1 do
    match s.[i] with
    | '\n' -> column := 0;
    | '\t' -> column := !column + 8;
    | _ -> incr column;
  done;
;;

let print_tag closing t =
  print (sprintf "<%s%s>" closing (string_of_tag t))
;;

let can_output () =
  match fifo_head () with
  | None -> false
  | Some (Opens { ordered = _; free = l }) when List.length l > 1 -> false
  | _ -> true
;;

let output t =
  fifo_put t;
  while can_output () do
    match fifo_get () with
    | Some (String s) -> print s;
    | Some (Opens {ordered = l1; free = l2}) ->
        List.iter (print_tag "") l2;
        List.iter (print_tag "") l1;
    | Some (Closes l) ->
        List.iter (print_tag "/") l;
    | None -> assert false
  done;
;;

let em_phase top st = stack_count EM top st mod 2 = 1;;

let close_partial top stack ctx =
  let rec loop good bad unsorted stack ctx =
    match unsorted with
    | [] -> top.ordered <- bad @ top.ordered;
            top.free <- good;
            output (Closes (List.rev bad));
    | e::t -> if is_compatible (e::good) stack ctx
              then loop (e::good) bad t stack ctx
              else loop good (e::bad) t stack ctx
  in
  loop [] [] top.free stack ctx
;;

let close_partial_sp top stack ctx =
  let rec loop good bad unsorted stack ctx =
    match unsorted with
    | [] -> top.ordered <- bad @ top.ordered;
            top.free <- good;
            output (Closes (List.rev bad));
    | e::t -> if is_compatible_sp (e::good) stack ctx
              then loop (e::good) bad t stack ctx
              else loop good (e::bad) t stack ctx
  in
  loop [] [] top.free stack ctx
;;

let close_all top =
  let lastfree = top.free in
  top.ordered <- top.free @ top.ordered;
  top.free <- [];
  output (Closes (List.rev lastfree));
;;

let rec clear_head stack =
  match stack with
  | {ordered = _; free = []} :: st -> clear_head st
  | _ -> stack
;;

let is_ok emflg szflg coflg st ctx =
  (emflg ==> (ctx.s || em_phase [] st = ctx.em))
  && (szflg ==> (stack_first_size [] st = ctx.size))
  && (coflg ==> (stack_first_color [] st = ctx.color))
;;

let is_better_or_same (e1, s1, c1) (e2, s2, c2) =
  (e2 ==> e1) && (s2 ==> s1) && (c2 ==> c1)
;;

let rec find_best st ctx emflg szflg coflg =
  let cur = if is_ok emflg szflg coflg st ctx
            then (emflg, szflg, coflg)
            else (false, false, false)
  in
  let deeper = match st with
    | [] -> (false, false, false)
    | { ordered = ord; free = [] } :: t ->
        find_best t ctx emflg szflg coflg
    | { ordered = ord; free = h::l } :: t ->
        let newst = { ordered = []; free = l } :: t in
        begin match h with
          | EM -> find_best newst ctx true szflg coflg
          | Size s -> find_best newst ctx emflg true coflg
          | Color c -> find_best newst ctx emflg szflg true
          | _ -> (false, false, false)
        end
  in
  if is_better_or_same cur deeper then cur else deeper
;;

let try_closing_em_size_color st ctx =
  let inite = (ctx.s || em_phase [] st = ctx.em) in
  let inits = (stack_first_size [] st = ctx.size) in
  let initc = (stack_first_color [] st = ctx.color) in
  let (be, bs, bc) = find_best st ctx inite inits initc in
  let rec loop st =
    if is_ok be bs bc st ctx then st else
    match st with
    | [] -> assert false
    | { ordered = _; free = [] } :: t -> loop t
    | ({ ordered = ord; free = h::l } as top) :: t ->
        top.ordered <- h::ord;
        top.free <- l;
        output (Closes [h]);
        loop st
  in
  loop st
;;

let is_ok_sp szflg coflg st ctx =
  (szflg ==> (stack_first_size [] st = ctx.sp_size))
  && (coflg ==> (ctx.sp_u > 0 && (stack_first_color [] st = ctx.sp_color)))
;;

let rec find_best_sp st ctx szflg coflg =
  let cur = if is_ok_sp szflg coflg st ctx
            then (false, szflg, coflg)
            else (false, false, false)
  in
  let deeper = match st with
    | [] -> (false, false, false)
    | { ordered = ord; free = [] } :: t ->
        find_best_sp t ctx szflg coflg
    | { ordered = ord; free = h::l } :: t ->
        let newst = { ordered = []; free = l } :: t in
        begin match h with
          | Size s -> find_best_sp newst ctx true coflg
          | Color c -> find_best_sp newst ctx szflg true
          | _ -> (false, false, false)
        end
  in
  if is_better_or_same cur deeper then cur else deeper
;;

let try_closing_size_color st ctx =
  let inits = (stack_first_size [] st = ctx.sp_size) in
  let initc = (stack_first_color [] st = ctx.sp_color) in
  let (_, bs, bc) = find_best_sp st ctx inits initc in
  let rec loop st =
    if is_ok_sp bs bc st ctx then st else
    match st with
    | [] -> assert false
    | { ordered = _; free = [] } :: t -> loop t
    | ({ ordered = ord; free = h::l } as top) :: t ->
        top.ordered <- h::ord;
        top.free <- l;
        output (Closes [h]);
        loop st
  in
  loop st
;;

let rec list_make e n =
  if n <= 0 then [] else e :: (list_make e (n-1))
;;

let add_missing stack ctx accu =
  let sctx = stack_to_ctx [] stack in
  let newtags = List.concat [
        (if sctx.size <> ctx.size then [Size ctx.size] else []);
        (if sctx.color <> ctx.color then [Color ctx.color] else []);
        (if ctx.s = false && sctx.em <> ctx.em then [EM] else []);
        (if ctx.b && not sctx.b then [B] else []);
        (if ctx.i && not sctx.i then [I] else []);
        (if ctx.s && not sctx.s then [S] else []);
        (if ctx.tt && not sctx.tt then [TT] else []);
        (list_make U (ctx.u - sctx.u));
      ]
  in
  let ops = { ordered = []; free = newtags } in
  output (Opens ops);
  ops :: stack
;;

let add_missing_sp stack ctx accu =
  let sctx = stack_to_ctx [] stack in
  let newtags = List.concat [
        (if sctx.size <> ctx.sp_size then [Size ctx.sp_size] else []);
        (if ctx.sp_u > 0 && sctx.color <> ctx.sp_color
           then [Color ctx.sp_color] else []);
        (if ctx.sp_tt && not sctx.tt then [TT] else []);
        (list_make U (ctx.sp_u - sctx.u));
      ]
  in
  let ops = { ordered = []; free = newtags } in
  output (Opens ops);
  ops :: stack
;;

let update_compatible stack ctx =
  let newstack = try_closing_em_size_color stack ctx in
  add_missing (clear_head stack) ctx []
;;

let update_compatible_sp stack spctx =
  let newstack = try_closing_size_color stack spctx in
  add_missing_sp (clear_head stack) spctx []
;;

let rec update_stack stack ctx =
  if is_compatible [] stack ctx
  then update_compatible stack ctx
  else if is_compatible [] (List.tl stack) ctx
  then begin
    close_partial (List.hd stack) (List.tl stack) ctx;
    update_compatible stack ctx
  end
  else begin
    close_all (List.hd stack);
    update_stack (List.tl stack) ctx
  end
;;

let rec update_stack_sp stack spctx =
  if is_compatible_sp [] stack spctx
  then update_compatible_sp stack spctx
  else if is_compatible_sp [] (List.tl stack) spctx
  then begin
    close_partial_sp (List.hd stack) (List.tl stack) spctx;
    update_compatible_sp (clear_head stack) spctx
  end
  else begin
    close_all (List.hd stack);
    update_stack_sp (List.tl stack) spctx
  end
;;

let same_spctx c1 c2 =
  c1.sp_tt = c2.sp_tt
  && c1.sp_size = c2.sp_size
  && c1.sp_u = c2.sp_u
  && (c1.sp_u = 0 || c1.sp_color = c2.sp_color)
;;

let rec optim prevstr prevspctx outstack ctx instack strm =
  match tok strm with
  | EOF -> List.iter close_all outstack;
  | Text s ->
      let newstack = update_stack outstack ctx in
      output (String s);
      optim s prevspctx newstack ctx instack strm;
  | Space s ->
      let newspctx = {sp_tt = ctx.tt; sp_size = ctx.size; sp_u = ctx.u;
                      sp_color = ctx.color}
      in
      if ctx.tt then begin
        let newstack = update_stack_sp outstack newspctx in
        output (String s);
        optim s newspctx newstack ctx instack strm;
      end else if prevstr = " " && same_spctx newspctx prevspctx then begin
        optim prevstr prevspctx outstack ctx instack strm;
      end else begin
        let newstack = update_stack_sp outstack newspctx in
        if !column > 50 then output (String "\n") else output (String "\n");
        optim " " newspctx newstack ctx instack strm;
      end
  | Open t ->
      let newctx = match t with
        | B -> {ctx with b = true}
        | EM -> {ctx with em = if ctx.s then ctx.em else not ctx.em}
        | I -> {ctx with i = true}
        | PL -> {plain with size = ctx.size; color = ctx.color}
        | S -> {ctx with s = true}
        | TT -> {ctx with tt = true}
        | U -> {ctx with u = if ctx.u >= 3 then 3 else 1 + ctx.u}
        | Size n -> {ctx with size = n}
        | Color c -> {ctx with color = c}
      in
      optim prevstr prevspctx outstack newctx (ctx::instack) strm;
   | Close t ->
      begin match instack with
        | [] -> error "unmatched closing tag"
        | h::t -> optim prevstr prevspctx outstack h t strm;
      end;
;;

let main inchan =
  optim "" plain_sp [] plain [] (Stream.of_channel inchan);
  flush stdout;
;;

main stdin;;
